/*
 * Copyright (C) 2005  Sven-Ola Tuecke  <sven-ola@gmx.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

/* Aug 27, 2005
 * 
 * This program calls an additional ioctl to set the BSSID in AdHoc
 * mode. Please note, that this BSSID will be overwritten immediately
 * if other stations are in range transmitting another BSSID if their
 * Timestamp (TSF) is older. You need a patched version of wl.o to
 * overcome this effect.
 */

/*
 * Feb 09, 2006
 *
 * Changed the buggy old program and added some options
 */
 
#include <stdio.h>
#include <unistd.h>
#include <stdint.h>
#include <string.h>
#include <getopt.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <wireless.h>
#include <wlioctl.h>
#include <wlutils.h>

#define WLC_R_REG 101
#define WLC_W_REG 102

#define BCM_SBF		0x120
#define BCM_TSFL 	0x180
#define BCM_TSFH 	0x184
#define BCM_BINT	0x188
#define BCM_TBTT	0x18c
#define SBF_TIME_UPDATE 0x10000000
	
#define lengthof(array) (int)(sizeof(array)/sizeof(*array))

/********* debug kernel mem (maybe for mmc_init) ***********

	int count = 0;
	unsigned long offs[32];
	unsigned long start = 0;
	unsigned char bssid[6] = {0x9e,0xd7,0x33,0x0c,0x33,0xc3-1};
	unsigned char newid[6] = {0xde,0xad,0xbe,0xef,0xba,0xbe};
	memset(offs, 0, sizeof(offs));
	
	start = PAGE_OFFSET;
	while(start < (unsigned long)high_memory - sizeof(bssid)) {
		if (0 == memcmp((unsigned char*)start, bssid, sizeof(bssid)))
		{
			if (count < lengthof(offs)) {
				printk("Self at %lx\n", start);
				offs[count] = start;
				count++;
			}
		}
		start++;
	}

	start = PAGE_OFFSET;
	bssid[5]++;
	while(start < (unsigned long)high_memory - sizeof(bssid)) {
		if (0 == memcmp((unsigned char*)start, bssid, sizeof(bssid)))
		{
			int i = 0;
			while(i < lengthof(offs) && offs[i] != start) i++;
			if (i < lengthof(offs)) {
				printk("Myown BSSID at %lx\n", start);
			}
			else {
				printk("Found BSSID at %lx\n", start);
#ifdef NIE
				memcpy(start, newid, sizeof(newid));
#endif
			}
		}
		start++;
	}

**************************************************************/

char* devname = 0;

void set_bcm(int off, unsigned long val) {
	rw_reg_t reg;
	reg.size = sizeof(reg.val);
	reg.byteoff = off;
	reg.val = val;
	wl_ioctl(devname, WLC_W_REG, &reg, sizeof(reg));
}

unsigned int get_bcm(int off) {
	rw_reg_t reg;
	reg.size = sizeof(reg.val);
	reg.byteoff = off;
	wl_ioctl(devname, WLC_R_REG, &reg, sizeof(reg));
	return reg.val;
}

int set_distance(char *v)
{
	rw_reg_t reg;
	uint32 shm;
	int val = 0;
	
	if (0 != v && 0 != *v) {
		val = atoi(v);
		val = 9+(val/150)+((val%150)?1:0);
		
		shm = 0x10;
		shm |= (val << 16);
		wl_ioctl(devname, 197, &shm, sizeof(shm));
		
		reg.byteoff = 0x684;
		reg.val = val + 510;
		reg.size = 2;
		wl_ioctl(devname, WLC_W_REG, &reg, sizeof(reg));
	}
	reg.byteoff = 0x684;
	reg.val = 0;
	reg.size = 2;
	wl_ioctl(devname, WLC_R_REG, &reg, sizeof(reg));
	return reg.val;
}

int main(int argc, char *argv[], char *environment[])
{
	char c;
	devname = argv[2];
	
	switch(c = getopt(argc, argv, "bdhqst")) {
	case 'b':
		if (3 < argc) {
			int i;
			unsigned int in[6];
			unsigned char bss[6];
			sscanf(argv[3], "%x:%x:%x:%x:%x:%x", &in[0], &in[1], &in[2], &in[3], &in[4], &in[5]);
			for(i = 0; i < lengthof(bss); i++) bss[i] = in[i] & 0xff;
			wl_ioctl(argv[2], WLC_SET_BSSID, bss, sizeof(bss));
			return 0;
		}
		break;
	case 'd':
		if (2 < argc) {
			printf("Distance: %d\n", set_distance(3 < argc?argv[3]:0));
			return 0;
		}
		break;
	case 'q':
		{
			FILE* f = fopen("/proc/net/dev", "r");
			if (0 == f) {
				perror("/proc/net/dev");
				return 1;
			}
			int skfd = socket(AF_INET, SOCK_DGRAM, 0);
			if (0 > skfd) {
				perror("No socket");
				return 1;
			}
			char s[256];
			while(0 != fgets(s, sizeof(s), f)) {
				char* p = s;
				while(0 != *p && p < s + sizeof(s) && ':' != *p) p++;
				if (0 != *p && p < s + sizeof(s)) {
					*p = 0;
					while(s < p && *(p - 1) != ' ' && *(p - 1) != '\t') p--;
					struct iwreq ifr;
					strncpy(ifr.ifr_name, p, sizeof(ifr.ifr_name - 1));
					int i;
					if (0 <= ioctl(skfd, SIOCGIWNAME, &ifr) || 
						(wl_ioctl(p, WLC_GET_MAGIC, &i, sizeof(i)) == 0 && 
						WLC_IOCTL_MAGIC == i))
					{
						/* Device has std wireless or is bcom */
						printf("%s\n", p);
						close(skfd);
						return 0;
					}
				}
			}
			close(skfd);
			fclose(f);
		}
		return 0;
	case 's':
		if (3 < argc) {
			unsigned long tsfh, tsfl;
			sscanf(argv[3], "%lx:%lx", &tsfh, &tsfl);
			
			set_bcm(BCM_SBF, get_bcm(BCM_SBF) | SBF_TIME_UPDATE);
			set_bcm(BCM_TSFH, tsfh);
			set_bcm(BCM_TSFL, tsfl);
			set_bcm(BCM_TBTT, tsfl + (get_bcm(BCM_BINT) >> 16) * 1000);
			set_bcm(BCM_SBF, get_bcm(BCM_SBF) & ~SBF_TIME_UPDATE);
			return 0;
		}
		break;
	case 't':
		if (2 < argc) {
			uint64_t tsf = get_bcm(BCM_TSFL) + ((uint64_t)get_bcm(BCM_TSFH) << 32);
			printf("%08llx:%08llx (%lld years, %lld days, %02lld:%02lld:%02lld)\n",
				(tsf & 0xffffffff00000000ULL) >> 32,
				(tsf & 0x00000000ffffffffULL) >> 0,
				(tsf / (365ULL * 24ULL * 60ULL * 60ULL * 1000000ULL)),
				(tsf / (24ULL * 60ULL * 60ULL * 1000000ULL)) % 365ULL,
				(tsf / (60ULL * 60ULL * 1000000ULL)) % 24ULL,
				(tsf / (60ULL * 1000000ULL)) % 60ULL,
				(tsf / (1000000ULL)) % 60ULL);
			return 0;
		}
		break;
	}
	fprintf(stderr, "Usage: %s -b [device] xx:xx:xx:xx:xx:xx to set the BSSID\n", argv[0]);
	fprintf(stderr, "       %s -d [device] (distance) in meter to set the ack timing\n", argv[0]);
	fprintf(stderr, "       %s -s [device] xxxxxxxx:yyyyyyyy to set the TSF\n", argv[0]);
	fprintf(stderr, "       %s -t [device] to print the TSF\n", argv[0]);
	fprintf(stderr, "       %s -q to print the wifi device (bcm or other)\n", argv[0]);
	return 1;
}
